@@ -20,6 +20,7 @@ package org.apache.felix.framework;
 
 import java.io.IOException;
 import java.lang.reflect.InvocationHandler;
+import java.lang.reflect.InvocationTargetException;
 import java.lang.reflect.Method;
 import java.lang.reflect.Proxy;
 import java.net.InetAddress;
@@ -58,10 +59,7 @@ public class URLHandlersStreamHandlerProxy extends URLStreamHandler
     implements URLStreamHandlerSetter, InvocationHandler
 {
     private static final Class[] URL_PROXY_CLASS;
-    private static final Object[] SERVICE_RANKING_PARAMS = new Object[]{"service.ranking"};
-    private static final Object[] SERVICE_ID_PARAMS = new Object[]{"service.id"};
     private static final Class[] STRING_TYPES = new Class[]{String.class};
-    private static final Class[] STRING_STRING_TYPES = new Class[]{String.class, String.class};
     private static final Method EQUALS;
     private static final Method GET_DEFAULT_PORT;
     private static final Method GET_HOST_ADDRESS;
@@ -128,28 +126,25 @@ public class URLHandlersStreamHandlerProxy extends URLStreamHandler
     private final SecureAction m_action;
     private final URLStreamHandler m_builtIn;
     private final URL m_builtInURL;
-    private final Object[] m_filter;
+    private final String m_protocol;
 
     public URLHandlersStreamHandlerProxy(String protocol, 
         SecureAction action, URLStreamHandler builtIn, URL builtInURL)
     {
+        m_protocol = protocol;
         m_service = null;
         m_action = action;
         m_builtIn = builtIn;
         m_builtInURL = builtInURL;
-        m_filter = new Object[]{"org.osgi.service.url.URLStreamHandlerService", 
-            "(&(objectClass=org.osgi.service.url.URLStreamHandlerService)(url.handler.protocol="
-            + protocol
-            + "))"};
     }
     
     private URLHandlersStreamHandlerProxy(Object service, SecureAction action)
     {
+        m_protocol = null;
         m_service = service;
         m_action = action;
         m_builtIn = null;
         m_builtInURL = null;
-        m_filter = null; 
     }
 
     //
@@ -296,6 +291,10 @@ public class URLHandlersStreamHandlerProxy extends URLStreamHandler
                     
                     throw new IOException("Extensions not supported or ambiguous context.");
                 }
+                catch (IOException ex)
+                {
+                    throw ex;
+                }
                 catch (Exception ex)
                 {
                     throw new IOException(ex.getMessage());
@@ -303,6 +302,10 @@ public class URLHandlersStreamHandlerProxy extends URLStreamHandler
             }
             return (URLConnection) OPEN_CONNECTION.invoke(svc, new Object[]{url});
         } 
+        catch (IOException ex)
+        {
+            throw ex;
+        }
         catch (Exception ex)  
         {
             throw new IllegalStateException("Stream handler unavailable due to: " + ex.getMessage());
@@ -331,9 +334,13 @@ public class URLHandlersStreamHandlerProxy extends URLStreamHandler
             {
                 m_action.setAccesssible(method);
                 return (URLConnection) method.invoke(svc, new Object[]{url, proxy});
-            } 
+            }
             catch (Exception e) 
             {
+                if (e instanceof IOException)
+                {
+                    throw (IOException) e;
+                }
                 throw new IOException(e.getMessage());
             }
         }
@@ -343,10 +350,18 @@ public class URLHandlersStreamHandlerProxy extends URLStreamHandler
         } 
         catch (Exception ex)  
         {
+            if (ex instanceof IOException)
+            {
+                throw (IOException) ex;
+            }
             throw new IllegalStateException("Stream handler unavailable due to: " + ex.getMessage());
         }
     }
 
+    // We use this thread local to detect whether we have a reentrant entry to the parseURL 
+    // method. This can happen do to some difference between gnu/classpath and sun jvms
+    // For more see inside the method.
+    private static final ThreadLocal m_loopCheck = new ThreadLocal();
     protected void parseURL(URL url, String spec, int start, int limit)
     {
         Object svc = getStreamHandlerService();
@@ -364,17 +379,50 @@ public class URLHandlersStreamHandlerProxy extends URLStreamHandler
             try 
             {
                 URL test = null;
+                // In order to cater for built-in urls being over-writable we need to use a 
+                // somewhat strange hack. We use a hidden feature inside the jdk which passes
+                // the handler of the url given as a context to a new URL to that URL as its
+                // handler. This way, we can create a new URL which will use the given built-in
+                // handler to parse the url. Subsequently, we can use the information from that
+                // URL to call set with the correct values. 
                 if (m_builtInURL != null)
                 {
-                    test = new URL(new URL(m_builtInURL, url.toExternalForm()), spec);
+                    // However, if we are on gnu/classpath we have to pass the handler directly
+                    // because the hidden feature is not there. Funnily, the workaround to pass
+                    // pass the handler directly doesn't work on sun as their handler detects
+                    // that it is not the same as the one inside the url and throws an exception
+                    // Luckily it doesn't do that on gnu/classpath. We detect that we need to 
+                    // pass the handler directly by using the m_loopCheck thread local to detect
+                    // that we parseURL has been called inside a call to parseURL.
+                    if (m_loopCheck.get() != null)
+                    {
+                        test = new URL(new URL(m_builtInURL, url.toExternalForm()), spec, (URLStreamHandler) svc);
+                    }
+                    else
+                    {
+                        // Set-up the thread local as we don't expect to be called again until we are
+                        // done. Otherwise, we are on gnu/classpath
+                        m_loopCheck.set(Thread.currentThread());
+                        try
+                        {
+                            test = new URL(new URL(m_builtInURL, url.toExternalForm()), spec);
+                        }
+                        finally
+                        {
+                            m_loopCheck.set(null);
+                        }
+                    }
                 }
                 else
                 {
+                    // We don't have a url with a built-in handler for this but still want to create
+                    // the url with the buil-in handler as we could find one now. This might not 
+                    // work for all handlers on sun but it is better then doing nothing. 
                     test = m_action.createURL(url, spec, (URLStreamHandler) svc);
                 }
-                    
+
                 super.setURL(url, test.getProtocol(), test.getHost(), test.getPort(),test.getAuthority(), 
-                   test.getUserInfo(), test.getPath(), test.getQuery(), test.getRef());
+                    test.getUserInfo(), test.getPath(), test.getQuery(), test.getRef());
             } 
             catch (Exception ex)  
             {
@@ -435,11 +483,70 @@ public class URLHandlersStreamHandlerProxy extends URLStreamHandler
         {
             return ((URLStreamHandlerService) svc).toExternalForm(url);
         }
-        try 
+        try
         {
-            return (String) TO_EXTERNAL_FORM.invoke( 
-                svc, new Object[]{url});
-        } 
+            try 
+            {
+                String result = (String) TO_EXTERNAL_FORM.invoke( 
+                    svc, new Object[]{url});
+                
+                // mika does return an invalid format if we have a url with the 
+                // protocol only (<proto>://null) - we catch this case now
+                if ((result != null) && (result.equals(url.getProtocol() + "://null")))
+                {
+                    result = url.getProtocol() + ":";
+                }
+                
+                return result;
+            }
+            catch (InvocationTargetException ex)
+            {
+               Throwable t = ex.getTargetException();
+               if (t instanceof Exception)
+               {
+                   throw (Exception) t;
+               }
+               else if (t instanceof Error)
+               {
+                   throw (Error) t;
+               }
+               else
+               {
+                   throw new IllegalStateException("Unknown throwable: " + t);
+               }
+            }
+        }
+        catch (NullPointerException ex)
+        {
+            // workaround for harmony and possibly J9. The issue is that
+            // their implementation of URLStreamHandler.toExternalForm() 
+            // assumes that URL.getFile() doesn't return null but in our
+            // case it can -- hence, we catch the NPE and do the work
+            // ourselvs. The only difference is that we check whether the
+            // URL.getFile() is null or not. 
+            StringBuffer answer = new StringBuffer();
+            answer.append(url.getProtocol());
+            answer.append(':');
+            String authority = url.getAuthority();
+            if ((authority != null) && (authority.length() > 0)) 
+            {
+                answer.append("//"); //$NON-NLS-1$
+                answer.append(url.getAuthority());
+            }
+
+            String file = url.getFile();
+            String ref = url.getRef();
+            if (file != null)
+            {
+                answer.append(file);
+            }
+            if (ref != null) 
+            {
+                answer.append('#');
+                answer.append(ref);
+            }
+            return answer.toString();
+        }
         catch (Exception ex)  
         {
             throw new IllegalStateException("Stream handler unavailable due to: " + ex.getMessage());
@@ -468,57 +575,19 @@ public class URLHandlersStreamHandlerProxy extends URLStreamHandler
             {
                 return m_builtIn;
             }
-            Object context = m_action.invoke(
-                m_action.getMethod(framework.getClass(), "getBundleContext", null),framework, null);
-            
-            Class contextClass = context.getClass();
-            
-            Object[] refs = (Object[]) m_action.invoke(
-                m_action.getMethod(contextClass, "getServiceReferences", STRING_STRING_TYPES), 
-                context, m_filter);
-            
-            Object ref = null;
-            int highestRank = -1;
-            long currentId = -1;
-            if (refs != null) 
-            {
-                // Loop through all service references and select the reference
-                // with the highest ranking and lower service identifier.
-                for (int i = 0; (refs != null) && (i < refs.length); i++)
-                {
-                    Class refClass = refs[i].getClass();
-                    Long idObj = (Long) m_action.invoke(m_action.getMethod(refClass, 
-                        "getProperty", STRING_TYPES), refs[i], SERVICE_ID_PARAMS);
-                    Integer rankObj = (Integer)  m_action.invoke(m_action.getMethod(refClass, 
-                        "getProperty", STRING_TYPES), refs[i], SERVICE_RANKING_PARAMS);
-                    // Ranking value defaults to zero.
-                    int rank = (rankObj == null) ? 0 : rankObj.intValue();
-                    if ((rank > highestRank) ||
-                        ((rank == highestRank) && (idObj.longValue() < currentId)))
-                    {
-                        ref = refs[i];
-                        highestRank = rank;
-                        currentId = idObj.longValue();
-                    }
-                }
-            }
+
             Object service = null;
-            if (ref != null) 
+            if (framework instanceof Felix)
             {
-                Class serviceRef = null;
-                Class[] interfaces = ref.getClass().getInterfaces();
-                for (int i = 0;i < interfaces.length; i++) 
-                {
-                    if ("org.osgi.framework.ServiceReference".equals(interfaces[i].getName()))
-                    {
-                        serviceRef = interfaces[i];
-                        break;
-                    }
-                }
-                service = m_action.invoke(
-                    m_action.getMethod(contextClass, "getService", new Class[]{serviceRef}), 
-                    context, new Object[]{ref});
+                service = ((Felix) framework).getStreamHandlerService(m_protocol);
             }
+            else
+            {
+                m_action.invoke(
+                    m_action.getMethod(framework.getClass(), "getStreamHandlerService", STRING_TYPES), 
+                    framework, new Object[]{m_protocol});
+            }
+
             if (service == null) 
             {
                 return m_builtIn;
@@ -538,6 +607,13 @@ public class URLHandlersStreamHandlerProxy extends URLStreamHandler
         }
         catch (Throwable t)
         {
+            // In case that we are inside tomcat - the problem is that the webapp classloader
+            // creates a new url to load a class. This gets us to this method. Now, if we 
+            // trigger a classload while executing tomcat is creating a new url and we end-up with
+            // a loop which is cut short after two iterations (because of a circularclassload). 
+            // We catch this exception (and all others) and just return the built-in handler
+            // (if we have any) as this way we at least eventually get started (this just means 
+            // that we don't use the potentially provided built-in handler overwrite). 
             return m_builtIn;
         }
     }
